import unittest

from zope.component import (
    adapts, getMultiAdapter, getSiteManager, getUtility, provideUtility)
from zope.component.interfaces import ComponentLookupError
from zope.interface import alsoProvides, Attribute, implements, Interface
from zope.publisher.interfaces.http import IHTTPRequest
from zope.schema import Int, Object, TextLine
from zope.security.checker import MultiChecker, ProxyFactory
from zope.security.management import endInteraction, newInteraction

from lazr.restful.declarations import (
    export_as_webservice_entry, exported, export_read_operation,
    export_write_operation, mutator_for, operation_for_version,
    operation_parameters)
from lazr.restful.interfaces import (
    IEntry, IResourceGETOperation, IWebServiceConfiguration,
    IWebServiceVersion)
from lazr.restful.marshallers import SimpleFieldMarshaller
from lazr.restful.metazcml import (
    AttemptToContributeToNonExportedInterface,
    ConflictInContributingInterfaces, find_interfaces_and_contributors)
from lazr.restful._resource import EntryAdapterUtility, EntryResource
from lazr.restful.testing.webservice import FakeRequest
from lazr.restful.testing.helpers import (
    create_test_module, TestWebServiceConfiguration, register_test_module)


class ContributingInterfacesTestCase(unittest.TestCase):
    """Tests for interfaces that contribute fields/operations to others."""

    def setUp(self):
        provideUtility(
            TestWebServiceConfiguration(), IWebServiceConfiguration)
        sm = getSiteManager()
        sm.registerUtility(
            ITestServiceRequestBeta, IWebServiceVersion, name='beta')
        sm.registerUtility(
            ITestServiceRequest10, IWebServiceVersion, name='1.0')
        sm.registerAdapter(ProductToHasBugsAdapter)
        sm.registerAdapter(ProjectToHasBugsAdapter)
        sm.registerAdapter(ProductToHasBranchesAdapter)
        sm.registerAdapter(DummyFieldMarshaller)
        self.beta_request = FakeRequest(version='beta')
        alsoProvides(
            self.beta_request, getUtility(IWebServiceVersion, name='beta'))
        self.one_zero_request = FakeRequest(version='1.0')
        alsoProvides(
            self.one_zero_request, getUtility(IWebServiceVersion, name='1.0'))
        self.product = Product()
        self.project = Project()

    def test_attributes(self):
        # The bug_count field comes from IHasBugs (which IProduct does not
        # provide, although it can be adapted into) but that field is
        # available in the webservice (IEntry) adapter for IProduct, and that
        # adapter knows it needs to adapt the product into an IHasBugs to
        # access .bug_count.
        self.product._bug_count = 10
        register_test_module('testmod', IProduct, IHasBugs)
        adapter = getMultiAdapter((self.product, self.beta_request), IEntry)
        self.assertEqual(adapter.bug_count, 10)

    def test_operations(self):
        # Although getBugsCount() is not provided by IProduct, it is available
        # on the webservice adapter as IHasBugs contributes it to IProduct.
        self.product._bug_count = 10
        register_test_module('testmod', IProduct, IHasBugs)
        adapter = getMultiAdapter(
            (self.product, self.beta_request),
            IResourceGETOperation, name='getBugsCount')
        self.assertEqual(adapter(), '10')

    def test_contributing_interface_with_differences_between_versions(self):
        # In the 'beta' version, IHasBranches.development_branches is exported
        # with its original name whereas for the '1.0' version it's exported
        # as 'development_branch_10'.
        self.product._dev_branch = Branch('A product branch')
        register_test_module('testmod', IProduct, IHasBranches)
        adapter = getMultiAdapter((self.product, self.beta_request), IEntry)
        self.assertEqual(adapter.development_branch, self.product._dev_branch)

        adapter = getMultiAdapter(
            (self.product, self.one_zero_request), IEntry)
        self.assertEqual(
            adapter.development_branch_10, self.product._dev_branch)

    def test_mutator_for_just_one_version(self):
        # On the 'beta' version, IHasBranches contributes a read only
        # development_branch field, but on version '1.0' that field can be
        # modified as we define a mutator for it.
        self.product._dev_branch = Branch('A product branch')
        register_test_module('testmod', IProduct, IHasBranches)
        adapter = getMultiAdapter((self.product, self.beta_request), IEntry)
        try:
            adapter.development_branch = None
        except AttributeError:
            pass
        else:
            self.fail('IHasBranches.development_branch should be read-only '
                      'on the beta version')

        adapter = getMultiAdapter(
            (self.product, self.one_zero_request), IEntry)
        self.assertEqual(
            adapter.development_branch_10, self.product._dev_branch)
        adapter.development_branch_10 = None
        self.assertEqual(
            adapter.development_branch_10, None)

    def test_contributing_to_multiple_interfaces(self):
        # Check that the webservice adapter for both IProduct and IProject
        # have the IHasBugs attributes, as that interface contributes to them.
        self.product._bug_count = 10
        self.project._bug_count = 100
        register_test_module('testmod', IProduct, IProject, IHasBugs)
        adapter = getMultiAdapter((self.product, self.beta_request), IEntry)
        self.assertEqual(adapter.bug_count, 10)

        adapter = getMultiAdapter((self.project, self.beta_request), IEntry)
        self.assertEqual(adapter.bug_count, 100)

    def test_multiple_contributing_interfaces(self):
        # Check that the webservice adapter for IProduct has the attributes
        # from both IHasBugs and IHasBranches.
        self.product._bug_count = 10
        self.product._dev_branch = Branch('A product branch')
        register_test_module('testmod', IProduct, IHasBugs, IHasBranches)
        adapter = getMultiAdapter((self.product, self.beta_request), IEntry)
        self.assertEqual(adapter.bug_count, 10)
        self.assertEqual(adapter.development_branch, self.product._dev_branch)

    def test_redacted_fields_for_empty_entry(self):
        # An entry which doesn't export any fields/operations will have no
        # redacted fields.
        class IEmpty(Interface):
            export_as_webservice_entry()
        class Empty:
            implements(IEmpty)
        register_test_module('testmod', IEmpty)
        entry_resource = EntryResource(Empty(), self.beta_request)
        self.assertEquals({}, entry_resource.entry._orig_interfaces)
        self.assertEquals([], entry_resource.redacted_fields)

    def test_redacted_fields_with_no_permission_checker(self):
        # When looking up an entry's redacted_fields, we take into account the
        # interface where the field is defined and adapt the context to that
        # interface before accessing that field.
        register_test_module('testmod', IProduct, IHasBugs, IHasBranches)
        entry_resource = EntryResource(self.product, self.beta_request)
        self.assertEquals([], entry_resource.redacted_fields)

    def test_redacted_fields_with_permission_checker(self):
        # When looking up an entry's redacted_fields for an object which is
        # security proxied, we use the security checker for the interface
        # where the field is defined.
        register_test_module('testmod', IProduct, IHasBugs, IHasBranches)
        newInteraction()
        try:
            secure_product = ProxyFactory(
                self.product,
                checker=MultiChecker([(IProduct, 'zope.Public')]))
            entry_resource = EntryResource(secure_product, self.beta_request)
            self.assertEquals([], entry_resource.redacted_fields)
        finally:
            endInteraction()

    def test_duplicate_contributed_attributes(self):
        # We do not allow a given attribute to be contributed to a given
        # interface by more than one contributing interface.
        testmod = create_test_module('testmod', IProduct, IHasBugs, IHasBugs2)
        self.assertRaises(
            ConflictInContributingInterfaces,
            find_interfaces_and_contributors, testmod)

    def test_contributing_interface_not_exported(self):
        # Contributing interfaces are not exported by themselves -- they only
        # contribute their exported fields/operations to other entries.
        class DummyHasBranches:
            implements(IHasBranches)
        dummy = DummyHasBranches()
        register_test_module('testmod', IProduct, IHasBranches)
        self.assertRaises(
            ComponentLookupError,
            getMultiAdapter, (dummy, self.beta_request), IEntry)

    def test_cannot_contribute_to_non_exported_interface(self):
        # A contributing interface can only contribute to exported interfaces.
        class INotExported(Interface):
            pass
        class IContributor(Interface):
            export_as_webservice_entry(contributes_to=[INotExported])
            title = exported(TextLine(title=u'The project title'))
        testmod = create_test_module('testmod', IContributor, INotExported)
        self.assertRaises(
            AttemptToContributeToNonExportedInterface,
            find_interfaces_and_contributors, testmod)

    def test_duplicate_contributed_methods(self):
        # We do not allow a given method to be contributed to a given
        # interface by more than one contributing interface.
        testmod = create_test_module('testmod', IProduct, IHasBugs, IHasBugs3)
        self.assertRaises(
            ConflictInContributingInterfaces,
            find_interfaces_and_contributors, testmod)

    def test_ConflictInContributingInterfaces(self):
        # The ConflictInContributingInterfaces exception states what are the
        # contributing interfaces that caused the conflict.
        e = ConflictInContributingInterfaces('foo', [IHasBugs, IHasBugs2])
        expected_msg = ("'foo' is exported in more than one contributing "
                        "interface: IHasBugs, IHasBugs2")
        self.assertEquals(str(e), expected_msg)

    def test_type_name(self):
        # Even though the generated adapters will contain stuff from various
        # different adapters, its type name is that of the main interface and
        # not one of its contributors.
        register_test_module('testmod', IProduct, IHasBugs)
        adapter = getMultiAdapter((self.product, self.beta_request), IEntry)
        self.assertEqual(
            'product', EntryAdapterUtility(adapter.__class__).singular_type)


class IProduct(Interface):
    export_as_webservice_entry()
    title = exported(TextLine(title=u'The product title'))
    # Need to define the two attributes below because we have a test which
    # wraps a Product object with a security proxy and later uses adapters
    # that access _dev_branch and _bug_count.
    _dev_branch = Attribute('dev branch')
    _bug_count = Attribute('bug count')


class Product(object):
    implements(IProduct)
    title = 'A product'
    _bug_count = 0
    _dev_branch = None


class IProject(Interface):
    export_as_webservice_entry()
    title = exported(TextLine(title=u'The project title'))


class Project(object):
    implements(IProject)
    title = 'A project'
    _bug_count = 0


class IHasBugs(Interface):
    export_as_webservice_entry(contributes_to=[IProduct, IProject])
    bug_count = exported(Int(title=u'Number of bugs'))
    not_exported = TextLine(title=u'Not exported')

    @export_read_operation()
    def getBugsCount():
        pass


class IHasBugs2(Interface):
    export_as_webservice_entry(contributes_to=[IProduct])
    bug_count = exported(Int(title=u'Number of bugs'))
    not_exported = TextLine(title=u'Not exported')


class IHasBugs3(Interface):
    export_as_webservice_entry(contributes_to=[IProduct])
    not_exported = TextLine(title=u'Not exported')

    @export_read_operation()
    def getBugsCount():
        pass


class ProductToHasBugsAdapter(object):
    adapts(IProduct)
    implements(IHasBugs)

    def __init__(self, context):
        self.context = context
        self.bug_count = context._bug_count

    def getBugsCount(self):
        return self.bug_count


class ProjectToHasBugsAdapter(ProductToHasBugsAdapter):
    adapts(IProject)


class IBranch(Interface):
    name = TextLine(title=u'The branch name')


class Branch(object):
    implements(IBranch)

    def __init__(self, name):
        self.name = name


class IHasBranches(Interface):
    export_as_webservice_entry(contributes_to=[IProduct])
    not_exported = TextLine(title=u'Not exported')
    development_branch = exported(
        Object(schema=IBranch, readonly=True),
        ('1.0', dict(exported_as='development_branch_10')),
        ('beta', dict(exported_as='development_branch')))

    @mutator_for(development_branch)
    @export_write_operation()
    @operation_parameters(value=TextLine())
    @operation_for_version('1.0')
    def set_dev_branch(value):
        pass


class ProductToHasBranchesAdapter(object):
    adapts(IProduct)
    implements(IHasBranches)

    def __init__(self, context):
        self.context = context

    @property
    def development_branch(self):
        return self.context._dev_branch

    def set_dev_branch(self, value):
        self.context._dev_branch = value


# One of our tests will try to unmarshall some entries, but even though we
# don't care about the unmarshalling itself, we need to register a generic
# marshaller so that the adapter lookup doesn't fail and cause an error on the
# test.
class DummyFieldMarshaller(SimpleFieldMarshaller):
    adapts(Interface, IHTTPRequest)


class ITestServiceRequestBeta(IWebServiceVersion):
    pass
class ITestServiceRequest10(IWebServiceVersion):
    pass
