# Copyright 2010 Canonical Ltd.  All rights reserved.

"""Test for lazr.restful.utils."""

__metaclass__ = type

import random
import unittest

from zope.publisher.browser import TestRequest
from zope.security.management import (
    endInteraction, newInteraction, queryInteraction)

from lazr.restful.utils import (get_current_browser_request,
    is_total_size_link_active, sorted_named_things)


class TestUtils(unittest.TestCase):

    def test_get_current_browser_request_no_interaction(self):
        # When there's no interaction setup, get_current_browser_request()
        # returns None.
        self.assertEquals(None, queryInteraction())
        self.assertEquals(None, get_current_browser_request())

    def test_get_current_browser_request(self):
        # When there is an interaction, it returns the interaction's request.
        request = TestRequest()
        newInteraction(request)
        self.assertEquals(request, get_current_browser_request())
        endInteraction()

    def test_is_total_size_link_active(self):
        # Parts of the code want to know if the sizes of collections should be
        # reported in an attribute or via a link back to the service.  The
        # is_total_size_link_active function takes the version of the API in
        # question and a web service configuration object and returns a
        # boolean that is true if a link should be used, false otherwise.

        # Here's the fake web service config we'll be using.
        class FakeConfig:
            active_versions = ['1.0', '2.0', '3.0']
            first_version_with_total_size_link = '2.0'

        # First, if the version is lower than the threshold for using links,
        # the result is false (i.e., links should not be used).
        self.assertEqual(is_total_size_link_active('1.0', FakeConfig), False)

        # However, if the requested version is equal to, or higher than the
        # threshold, the result is true (i.e., links should be used).
        self.assertEqual(is_total_size_link_active('2.0', FakeConfig), True)
        self.assertEqual(is_total_size_link_active('3.0', FakeConfig), True)

    def test_name_sorter(self):
        # The WADL generation often sorts classes or functions by name; the
        # sorted_named_things helper... helps.

        # First we need some named things to sort.
        class Thing:
            pass

        def make_named_thing():
            thing = Thing()
            thing.__name__ = random.choice('abcdefghijk')
            return thing

        # The function sorts on the object's __name__ attribute, which
        # functions and classes have, see:
        assert hasattr(Thing, '__name__')
        assert hasattr(make_named_thing, '__name__')

        # Now we can make a bunch of things with randomly ordered names and
        # show that sorting them does order them by name.
        things = sorted_named_things(make_named_thing() for i in range(10))
        names = [thing.__name__ for thing in things]
        self.assertEqual(names, sorted(names))

    # For the sake of convenience, test_get_current_web_service_request()
    # and tag_request_with_version_name() are tested in test_webservice.py.

